package algo.datastructure;

import algo.branch.Packing;
import algo.util.InitialGraph;
import algo.util.InitialVertex;

import java.util.*;

public class Graph implements Iterable<Vertex> {

    public final int initialVertexCount;
    private final PartSetVertices vertices;
    public int vertexCount;

    private Edge[][] edges = null;

    public NCs ncs;

    public boolean updateNCs = false;
    public boolean updateNonCommons = false;

    public Graph(InitialGraph initialGraph) {
        vertexCount = 0;

        initialVertexCount = initialGraph.vertices.size();
        vertices = new PartSetVertices(initialVertexCount);

        List<InitialVertex> initialVertices = initialGraph.vertices;

        for (InitialVertex initialVertex : initialVertices)
            this.addVertex(new Vertex(initialVertex, initialVertexCount));

        initializeEdges(initialGraph);
    }

    public void addVertex(Vertex v) {
        vertices.add(v);
        vertexCount++;
    }

    public void removeVertex(Vertex v) {
        vertices.remove(v);
        vertexCount--;
    }

    public void initializeNCsDataStructure() {
        ncs = new NCs(this);
        updateNCs = true;
    }

    public void editEdge(Edge edge) {
        boolean oldIsNotMergeAble = edge.isNotMergeableWithLabel();

        if (edge.isEdge) removeEdge(edge);
        else addEdge(edge);
        edge.isEdited = !edge.isEdited;

        boolean newIsNotMergeAble = edge.isNotMergeableWithLabel();
        if (oldIsNotMergeAble != newIsNotMergeAble)
            Packing.updatePackingByUpdateNotMergeAble(edge, newIsNotMergeAble);

        if (updateNonCommons) edge.updateCommonsAndNonCommonsByEdit(this);
        if (updateNCs) updateNCsByEdit(edge);
    }

    public Vertex get(int i) {
        return vertices.get(i);
    }

    public void addEdge(Edge edge) {
        Vertex a = edge.u;
        Vertex b = edge.v;
        a.neighbors.add(b);
        b.neighbors.add(a);
        edge.isEdge = true;
    }

    public void removeEdge(Edge edge) {
        Vertex a = edge.u;
        Vertex b = edge.v;
        a.neighbors.remove(b);
        b.neighbors.remove(a);
        edge.isEdge = false;
    }

    public void addEdgeOneDirectional(Vertex u, Vertex v) {
        u.neighbors.add(v);
    }

    public void removeEdgeOneDirectional(Vertex u, Vertex v) {
        u.neighbors.remove(v);
    }

    private void initializeEdges(InitialGraph initialGraph) {
        edges = new Edge[vertexCount][vertexCount];
        //create Edge-objects
        for (int i = 0; i < vertexCount; i++) {
            Vertex u = get(i);
            for (int j = i + 1; j < vertexCount; j++) {
                Vertex v = get(j);
                Edge e = new Edge(u, v, this);
                edges[i][j] = e;
                edges[j][i] = e;
            }
        }
        //add neighbors
        for (int i = 0; i < initialGraph.vertices.size(); i++) {
            Vertex u = get(i);
            InitialVertex initialVertex = initialGraph.vertices.get(i);
            for (InitialVertex initNeighbor : initialVertex.neighbors) {
                if (u.id < initNeighbor.id) {
                    Vertex v = get(initNeighbor.id);
                    Edge edge = getEdge(u, v);

                    editEdge(edge);
                    edge.isEdge = true;
                    edge.isBlocked = false;
                    edge.isEdited = false;
                }
            }
        }
    }

    public void initializeNonCommons() {
        updateNonCommons = true;
        for (int i = 0; i < vertexCount; i++) {
            Vertex u = get(i);
            for (int j = i + 1; j < vertexCount; j++) {
                Vertex v = get(j);
                Edge e = getEdge(u, v);
                for (Vertex potNonCommon : vertices) {
                    if (potNonCommon == u || potNonCommon == v) continue;
                    if (areNeighbors(u, potNonCommon) != areNeighbors(v, potNonCommon)) {
                        e.addNonCommon(potNonCommon);
                    }
                }
            }
        }
    }

    public boolean areNeighbors(Vertex u, Vertex v) {
        return getEdge(u, v).isEdge;
    }

    public Edge getEdge(Vertex u, Vertex v) {
        return edges[u.id][v.id];
    }

    public int computeEdgeCount() {
        return vertices.stream().map(Vertex::degree).reduce(0, Integer::sum) / 2;
    }

    public List<Vertex> getBranchVertices(int t) {
        List<Vertex> branchVertices = new ArrayList<>(t + 1);
        for (Vertex vertex : this) {
            boolean newNC = true;
            for (Vertex branchVertex : branchVertices) {
                if (areSimilar(vertex, branchVertex)) {
                    newNC = false;
                    break;
                }
            }
            if (newNC) branchVertices.add(vertex);
            if (branchVertices.size() > t) break;
        }
        return branchVertices;
    }

    public boolean areSimilar(Vertex u, Vertex v) {
        return getEdge(u, v).getNumberOfNonCommons() == 0 && haveSameAdjacencyInsideAndTowards(u, v);
    }

    public boolean haveSameAdjacencyInsideAndTowards(Vertex u, Vertex v) {
        Label con1 = u.getInsideAdjacency();
        Label con2 = v.getInsideAdjacency();
        Label connectionOfEdge = (getEdge(u, v).isEdge) ? Label.ADJACENT : Label.SEPARATE;
        if (con1 == con2) {
            if (con1 == connectionOfEdge) return true;
            if (con1 == Label.UNDECIDED && u.weight == 1 && v.weight == 1)
                return true;
        }
        return con1 == Label.UNDECIDED && u.weight == 1 && connectionOfEdge == con2
                || con2 == Label.UNDECIDED && v.weight == 1 && connectionOfEdge == con1;
    }

    public void updateNCsByEdit(Edge e) {
        ncs.updateNCsByEdit(e);
    }

    public int ncCount() {
        return ncs.ncCount;
    }

    public void searchNCAndMoveVertex(Vertex v) {
        ncs.searchNCAndMoveVertex(v);
    }

    public NC getNC(int i) {
        return ncs.ncs[i];
    }

    public void selectFirstVertices(int numberToSelect, Comparator<Vertex> comparator) {
        if (numberToSelect >= vertexCount) return;
        vertices.selectFirstVertices(numberToSelect, comparator);
    }

    @Override
    public Iterator<Vertex> iterator() {
        return vertices.iterator();
    }

    public boolean inSameNC(Vertex u, Vertex v) {
        return u.currentNC == v.currentNC;
    }
}