package algo.heuristic;

import algo.datastructure.*;
import algo.util.Util;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;

public class LocalSearch extends Util {

    public static int improve(Graph g, BlockPartition bp, List<NC> ncs) {
        Block auxBlock = bp.addEmptyBlock(ncs.size());
        bp.initializeUseOfEdgesFromVertexToBlock(g);
        boolean progress = true;
        while (progress) {
            progress = improveByMovingVertexToBlock(g, bp, auxBlock);
            progress = improveBySwapVertices(g, bp) || progress;
            progress = improveByMovingNeighborhoodToBlock(bp, auxBlock) || progress;
        }
        bp.removeEmptyBlock(auxBlock);
        return bp.computeCost();
    }

    private static boolean improveByMovingVertexToBlock(Graph g, BlockPartition bp, Block auxBlock) {
        boolean progress = false;
        for (Vertex vertex : g) {
            Block originBlock = vertex.currentBlock;
            bp.moveVertexToBlock(vertex, auxBlock);
            Set<Block> bestBlocks = bp.computeBestBlocksByMovingVertex(vertex).bestBlocks;
            progress = progress || !bestBlocks.contains(originBlock);

            Block newBlock = getRandomElementOf(bestBlocks);
            bp.moveVertexToBlock(vertex, newBlock);
        }
        return progress;
    }

    private static boolean improveBySwapVertices(Graph g, BlockPartition bp) {
        boolean progress = false;
        for (int i = 0; i < g.vertexCount; i++) {
            Vertex u = g.get(i);
            for (int j = i + 1; j < g.vertexCount; j++) {
                Vertex v = g.get(j);
                Block bU = u.currentBlock;
                Block bV = v.currentBlock;
                if (bU == bV) continue;

                int costIncrease = getCostIncreaseBySwap(g, bp, u, v);
                if (costIncrease < 0) {
                    bp.moveVertexToBlock(u, bV);
                    bp.moveVertexToBlock(v, bU);
                    progress = true;
                }
            }
        }
        return progress;
    }

    private static int getCostIncreaseBySwap(Graph g, BlockPartition bp, Vertex u, Vertex v) {
        Block bU = u.currentBlock;
        Block bV = v.currentBlock;

        int costIncrease = 0;
        //costIncrease to other blocks
        for (Block block : bp.getBlocks()) {
            if (block == bU || block == bV) continue;
            int edgesFromUToBlock = bp.getEdgeCountBetween(u, block);
            int edgesFromVToBlock = bp.getEdgeCountBetween(v, block);
            if (edgesFromUToBlock == edgesFromVToBlock) continue;

            int newEdgesCountBU = bp.getEdgeCountBetween(bU, block)
                    - edgesFromUToBlock + edgesFromVToBlock;

            int newEdgesCountBV = bp.getEdgeCountBetween(bV, block)
                    + edgesFromUToBlock - edgesFromVToBlock;

            int currentCost = bp.getCostBetween(bU, block) + bp.getCostBetween(bV, block);
            int newCost = Math.min(block.size * bU.size - newEdgesCountBU, newEdgesCountBU)
                    + Math.min(block.size * bV.size - newEdgesCountBV, newEdgesCountBV);
            costIncrease += (newCost - currentCost);
        }
        //within
        int currentCost = bp.getCostWithin(bU) + bp.getCostWithin(bV);
        int newEdgesWithinBU = bp.getEdgeCountBetween(bU, bU)
                - bp.getEdgeCountBetween(u, bU) + bp.getEdgeCountBetween(v, bU);
        int newEdgesWithinBV = bp.getEdgeCountBetween(bV, bV)
                + bp.getEdgeCountBetween(u, bV) - bp.getEdgeCountBetween(v, bV);
        if (g.areNeighbors(u, v)) {
            newEdgesWithinBU -= 1;
            newEdgesWithinBV -= 1;
        }
        int newCost = Math.min(over2(bU.size) - newEdgesWithinBU, newEdgesWithinBU)
                + Math.min(over2(bV.size) - newEdgesWithinBV, newEdgesWithinBV);
        costIncrease += (newCost - currentCost);

        //between bU and bV
        int currentCostBetween = bp.getCostBetween(bU, bV);
        int newEdgesBetween = bp.getEdgeCountBetween(bU, bV)
                + bp.getEdgeCountBetween(u, bU) + bp.getEdgeCountBetween(v, bV)
                - bp.getEdgeCountBetween(v, bU) - bp.getEdgeCountBetween(u, bV);
        if (g.areNeighbors(u, v)) {
            newEdgesBetween += 2;
        }
        int newCostBetween = Math.min(bU.size * bV.size - newEdgesBetween, newEdgesBetween);
        costIncrease += (newCostBetween - currentCostBetween);

        return costIncrease;
    }

    private static boolean improveByMovingNeighborhoodToBlock(BlockPartition bp, Block auxBlock) {
        boolean progress = false;
        for (Block block : bp.getBlocks()) {
            List<List<Vertex>> listOfNCsAsList = getNCDistributionAsList(block);
            //compute cost for this block
            int costOfBlock = bp.computeCostOfBlock(block);
            for (Block blockToAugment : bp.getBlocks()) {
                if (blockToAugment == auxBlock || blockToAugment == block) continue;

                int costOfOtherBlock = bp.computeCostOfBlock(blockToAugment);
                int oldCost = costOfBlock - costOfOtherBlock - bp.getCostBetween(block, blockToAugment);
                for (List<Vertex> verticesToMove : listOfNCsAsList)
                    verticesToMove.forEach(v -> bp.moveVertexToBlock(v, blockToAugment));

                int newCost = bp.computeCostOfBlock(block)
                        + bp.computeCostOfBlock(blockToAugment)
                        - bp.getCostBetween(block, blockToAugment);
                if (newCost < oldCost) {
                    progress = true;
                    break;
                } else {
                    for (List<Vertex> verticesToMove : listOfNCsAsList)
                        verticesToMove.forEach(v -> bp.moveVertexToBlock(v, block));
                }
            }
        }
        return progress;
    }

    private static List<List<Vertex>> getNCDistributionAsList(Block block) {
        List<List<Vertex>> listOfNCsAsList = new LinkedList<>();
        for (int i = 0; i < block.size; i++) {
            Vertex unsortedVertex = block.get(i);
            boolean putInNCList = false;
            for (List<Vertex> verticesOfNC : listOfNCsAsList) {
                Vertex vertexOfNC = verticesOfNC.get(0);
                if (unsortedVertex.initialNC == vertexOfNC.initialNC) {
                    verticesOfNC.add(unsortedVertex);
                    putInNCList = true;
                    break;
                }
            }

            if (!putInNCList) {//add new list
                List<Vertex> newNCList = new LinkedList<>();
                listOfNCsAsList.add(newNCList);
                newNCList.add(unsortedVertex);
            }
        }
        return listOfNCsAsList;
    }
}
